import os

from flask import Flask
from flask_login import LoginManager

from .database import db, db_migrator
from .ldap_authentication.authenticator import DocLdapAuthenticator
from .messages import LOGIN_MANAGER_MESSAGE
from .mocks.ldap_authentication import DummyLdapAuthenticator

configuration_switch = {
    "dev": f"{__name__}.config.DevConfig",  # Development configuration (fake LDAP)
    "staging": f"{__name__}.config.StagingConfig",  # Staging configuration (should be as close as possible to prod)
    "production": f"{__name__}.config.ProductionConfig",  # Production configuration
}

ENV = os.environ.get("ENV", "dev")

# SET UP =====================================

login_manager = LoginManager()
login_manager.login_view = "auth.login"
login_manager.login_message = LOGIN_MANAGER_MESSAGE


def get_ldap_handler(env):
    if env == "dev":
        return DummyLdapAuthenticator()
    return DocLdapAuthenticator()


ldap_service = get_ldap_handler(ENV)


# ===================================================


def create_app(test_configuration=None):
    """Application factory method"""
    app = Flask(__name__)

    # Configure the application
    if test_configuration:
        app.config.update(test_configuration)
    else:
        app.config.from_object(configuration_switch[ENV])

    # Register extensions ################################
    login_manager.init_app(app)
    db.init_app(app)
    with app.app_context():
        db_migrator.init_app(
            app,
            db,
            compare_type=True,
            render_as_batch=db.engine.url.drivername == "sqlite",
        )

    # Register CLI commands ####################################
    from .cli import drop_all, populate_dev

    with app.app_context():
        app.cli.add_command(populate_dev)
        app.cli.add_command(drop_all)

    # Register blueprints ####################################
    from .views import auth, home, staff, student

    app.register_blueprint(auth.bp)
    app.register_blueprint(home.bp)
    app.register_blueprint(staff.bp)
    app.register_blueprint(student.bp)

    return app
