from flask import Blueprint, flash, redirect, render_template, request, url_for
from flask_login import current_user, login_required
from sqlalchemy import or_

from app import messages
from app.database import db
from app.models.person import Person
from app.models.project import Project
from app.models.shortlist import Shortlisting
from app.utils.decorators import only_before_deadline_for_students, students_only

bp = Blueprint("student", __name__, url_prefix="/student")


@bp.route("/projects")
@login_required
@students_only
def projects():
    active_projects: list[Project] = (
        Project.query.filter(
            or_(Project.on_behalf.is_(None), Project.on_behalf == current_user.username)
        )
        .filter(Project.deleted.is_(None))
        .all()
    )
    short_listings = Shortlisting.query.filter_by(student=current_user.username).all()
    short_listings = {s.project_id for s in short_listings}
    return render_template(
        "pages/student/projects.html",
        projects=active_projects,
        short_listings=short_listings,
    )


@bp.route("/projects/shortlisted")
@login_required
@students_only
def projects_ranking():
    shortlisted_projects: list[Project] = (
        Project.query.join(Shortlisting)
        .filter(Shortlisting.student == current_user.username)
        .order_by(Shortlisting.ranking)
        .all()
    )

    full_names = {}

    for shortlisting in shortlisted_projects:
        username = (
            shortlisting.proposer
            if shortlisting.on_behalf is None
            else shortlisting.on_behalf
        )

        person = Person.query.filter(Person.username == username).first() or Person(
            username=username, firstname="", lastname=""
        )

        full_names[shortlisting.id] = (
            person.full_name if person.full_name else person.username
        )

    return render_template(
        "pages/student/shortlisted-projects.html",
        projects=shortlisted_projects,
        full_names=full_names,
    )


@bp.route("/projects/rankings", methods=["PUT"])
@login_required
@students_only
@only_before_deadline_for_students
def update_ranking():
    project_ids = request.json
    current_shortlist: list[Shortlisting] = Shortlisting.query.filter(
        Shortlisting.project_id.in_(project_ids)
    ).filter_by(student=current_user.username)
    for shortlisting in current_shortlist:
        shortlisting.ranking = project_ids.index(shortlisting.project_id) + 1
    db.session.commit()
    return "", 204


@bp.route("/projects/<project_id>")
@login_required
@students_only
def view_project(project_id):
    if project := Project.query.get(project_id):
        username = project.proposer if project.on_behalf is None else project.on_behalf

        person = Person.query.filter(Person.username == username).first() or Person(
            username=username, firstname="", lastname=""
        )

        return render_template(
            "pages/student/project-view.html", project=project, person=person
        )
    flash(messages.PROJECT_NOT_FOUND)
    return redirect(url_for("student.projects"))


@bp.route("/projects/<project_id>/shortlist")
@login_required
@students_only
@only_before_deadline_for_students
def shortlist_project(project_id):
    if Project.query.get(project_id):
        shortlisting = Shortlisting(
            student=current_user.username, project_id=project_id
        )
        db.session.add(shortlisting)
        db.session.commit()
    else:
        flash(messages.PROJECT_NOT_FOUND)
    return redirect(url_for("student.projects"))


@bp.route("/projects/<project_id>/unshortlist")
@login_required
@students_only
@only_before_deadline_for_students
def unshortlist_project(project_id):
    if Project.query.get(project_id):
        shortlist: Shortlisting = (
            Shortlisting.query.filter(Shortlisting.project_id == project_id)
            .filter_by(student=current_user.username)
            .first()
        )
        if shortlist:
            db.session.delete(shortlist)
            db.session.commit()
        else:
            flash(messages.SHORTLIST_NOT_FOUND)
    else:
        flash(messages.PROJECT_NOT_FOUND)
    return redirect(url_for("student.projects"))
